class Coordinate:
    __slots__ = ("x", "y")

    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __add__(self, other):
        return Coordinate(self.x + other.x, self.y + other.y)

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y



import curses


class Key:
    def __init__(self, *values):
        self.values = values
        self._hash = hash(values)
        self._keyset = set(values)

    def __eq__(self, other):
        return self._hash == other._hash

    def __hash__(self):
        return self._hash


class Keys:
    ESC = Key(27)
    TAB = Key(ord("\t"), ord("n"))
    SHIFT_TAB = Key(353, ord("N"))
    VISUAL = Key(ord("v"), ord("V"))
    COPY = Key(ord("c"), ord("y"))
    QUIT = Key(ord("q"))
    UP = Key(curses.KEY_UP, ord("k"))
    DOWN = Key(curses.KEY_DOWN, ord("j"))
    LEFT = Key(curses.KEY_LEFT, ord("h"))
    RIGHT = Key(curses.KEY_RIGHT, ord("l"))
    HELP = Key(ord("?"))
    ALL = [ESC, TAB, SHIFT_TAB, VISUAL, COPY, QUIT, UP, DOWN, LEFT, RIGHT, HELP]
    _id_to_key = {id: key for key in ALL for id in key.values}

    @staticmethod
    def to_key(key: int) -> Key:
        return Keys._id_to_key.get(key)



import curses
import math
import numpy as np


class C:
    WHITE = curses.COLOR_WHITE
    BLACK = curses.COLOR_BLACK
    BLUE = curses.COLOR_BLUE
    GREEN = curses.COLOR_GREEN
    YELLOW = curses.COLOR_YELLOW
    MAGENTA = curses.COLOR_MAGENTA
    LIGHT_MAGENTA = 213


class Styles:
    def __init__(self, cellwidth: int, precision: int):
        self.separator: str = "|"
        self.precision: int = precision
        self.init_colors()
        self.width: int = cellwidth
        self.lnwidth: int = None
        self.cell_fmt: str = None
        self.lineno_fmt: str = None

    def init(self, height: int):
        self.lnwidth: int = int(math.log10(height) + 2)
        self.cell_fmt: str = self.get_cell_format_string()
        self.lineno_fmt: str = self.get_lineno_format_string()

    def init_colors(self):
        curses.init_pair(1, C.WHITE, C.BLUE)
        curses.init_pair(2, C.MAGENTA, C.BLACK)
        curses.init_pair(3, C.BLUE, C.BLACK)
        curses.init_pair(4, C.GREEN, C.BLACK)
        curses.init_pair(5, C.BLACK, C.YELLOW)
        curses.init_pair(6, C.LIGHT_MAGENTA, C.BLACK)
        self.c_selection = curses.color_pair(1)
        self.c_sheets = curses.color_pair(2)
        self.c_border = curses.color_pair(3)
        self.c_sheet_selection = curses.A_UNDERLINE | curses.color_pair(6)
        self.c_header = curses.A_UNDERLINE | self.c_border
        self.c_footer = curses.A_UNDERLINE | curses.A_BOLD | curses.color_pair(4)
        self.c_lineno = self.c_border
        self.c_visual = curses.color_pair(5)

    def get_cell_format_string(self) -> str:
        return "{:>" + f"{self.width}" + "}" + self.separator

    def get_lineno_format_string(self) -> str:
        return "{:>" + f"{self.lnwidth - 1}" + "}" + self.separator

    def get_header_format_string(self, num_cols: int) -> str:
        return ("{:" f"{self.width}" + "}" + self.separator) * num_cols

    def format_footer(self, string: str, width) -> str:
        if len(string) > width:
            string = string[: width - 3] + "..."
        return ("{:>" + f"{width}" + "}").format(string)

    def format_cell(self, x) -> str:
        if isinstance(x, (float, np.floating)):
            x = round(x, self.precision)
        x = str(x)
        if len(x) > self.width:
            x = x[: self.width - 3] + "..."
        return self.cell_fmt.format(x)



#!/usr/bin/env python

import curses
from argparse import ArgumentParser

from .grid import Grid
from .styles import Styles
from .utils import read_spreadsheet, InvalidFileException


def run(scr, sheets, args):
    grid = Grid(
        scr,
        sheets,
        Styles(args.cellwidth, args.precision),
    )
    while grid.handle_press(scr.getch()):
        pass


def main():
    parser = ArgumentParser()
    parser.add_argument(
        "file",
        help="Spreadsheet path (csv, xls, xlsx, xlsm, xlsb, odf, ods or odt)",
        metavar="FILE",
    )
    parser.add_argument(
        "-d",
        "--delimiter",
        help="Delimiter for csv files",
        metavar="DELIMITER",
        default=",",
    )
    parser.add_argument(
        "-c",
        "--cellwidth",
        help="Width of a cell",
        metavar="CELLWIDTH",
        type=int,
        default=10,
    )
    parser.add_argument(
        "-p",
        "--precision",
        help="Precision of floating point numbers",
        metavar="PRECISION",
        type=int,
        default=2,
    )
    parser.add_argument(
        "--fillna",
        help="Value to use to fill holes in the spreadsheet",
        metavar="FILLNA",
        default=None,
    )
    parser.add_argument(
        "--encoding",
        help="Encoding for CSV files",
        metavar="ENCODING",
        default="utf-8",
    )
    try:
        args = parser.parse_args()
        sheets = read_spreadsheet(args.file, args.delimiter, args.fillna, args.encoding)
    except InvalidFileException as e:
        print(e)
        exit()
    except TypeError as e:
        print(e)
        parser.print_usage()
        exit()
    except FileNotFoundError:
        print("Cannot find or open {}".format(args.file))
        exit()
    curses.wrapper(run, sheets, args)



from setuptools import setup

requirements = [
    "openpyxl",
    "odfpy",
    "pyperclip",
    "numpy",
]

with open("README.md") as f:
    long_description = f.read()

setup(
    name="texel",
    packages=["texel"],
    version="0.2.0",
    description="Command line interface for reading spreadsheets inside terminal.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    python_requires=">=3",
    entry_points={"console_scripts": ["texel = texel.texel:main"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    author="Lauri Tuominen",
    author_email="lauri.a.tuominen@gmail.com",
    url="https://github.com/lauriat/texel",
)



import ast
import csv
import os
from numbers import Number
from typing import Any, Dict, Union


import numpy as np
from openpyxl import load_workbook as load_excel
from odf.opendocument import load as load_odf
from odf.table import Table, TableCell, TableRow


EXCEL_FORMATS = ("xlsx", "xlsm", "xltx", "xltm")
ODF_FORMATS = ("odf", "odt", "ods")
SUPPORTED_FORMATS = ("csv", *ODF_FORMATS, *EXCEL_FORMATS)


class InvalidFileException(Exception):
    pass


class SpreadsheetReader:
    def __init__(self, filename: str, delimiter: str, fillna: Any, encoding: str):
        self.filename = filename
        self.delimiter = delimiter
        self.fillna = fillna
        self.encoding = encoding
        self.ft = self.filename.split(".")[-1]

    def read(self) -> Dict[str, np.ndarray]:
        self._validate()
        if self.ft in EXCEL_FORMATS:
            sheetdict = self._read_excel()
        elif self.ft in ODF_FORMATS:
            sheetdict = self._read_odf()
        else:
            sheetdict = self._read_csv()
        if self.fillna is not None:
            for key in sheetdict:
                sheetdict[key][sheetdict[key] == np.nan] = self.fillna
        return sheetdict

    def _validate(self):
        self._check_exists()
        self._check_format()

    def _check_format(self):
        if self.ft not in SUPPORTED_FORMATS:
            raise InvalidFileException(
                f"Invalid file format ({self.ft})\n"
                + f"Supported formats are: {', '.join(SUPPORTED_FORMATS)}"
            )

    def _check_exists(self):
        if not os.path.exists(self.filename):
            raise FileNotFoundError()

    def _parse_cell(self, value: Union[str, Number, TableCell]) -> Union[str, Number]:
        if isinstance(value, Number):
            return value
        if value is None:
            return np.nan
        value = str(value).strip()
        if len(value) > 0:
            try:
                return ast.literal_eval(value)
            except (SyntaxError, ValueError):
                return value
        return np.nan

    def _parse_odfcell(self, cell: TableCell):
        if cell.getAttribute("valuetype") != "string":
            cell = cell.getAttribute("value")
        return self._parse_cell(cell)

    def _read_csv(self):
        lines = []
        with open(self.filename, encoding=self.encoding) as csvfile:
            for line in csv.reader(csvfile, delimiter=self.delimiter):
                lines.append(list(map(self._parse_cell, line)))
        return {self.filename: self._to_array(lines)}

    def _read_excel(self):
        sheets = {}
        wb = load_excel(filename=self.filename, data_only=True)
        for sheet in wb.sheetnames:
            lines = [[self._parse_cell(e.value) for e in row] for row in wb[sheet].rows]
            sheets[sheet] = self._to_array(lines)
        return sheets

    def _read_odf(self):
        sheets = {}
        wb = load_odf(self.filename)
        for sheet in wb.getElementsByType(Table):
            rows = [
                list(map(self._parse_odfcell, row.getElementsByType(TableCell)))
                for row in sheet.getElementsByType(TableRow)
            ]
            sheets[sheet.getAttribute("name")] = self._to_array(rows)
        return sheets

    def _to_array(self, lines):
        shape = len(lines), max(map(len, lines))
        arr = np.zeros(shape, dtype=object)
        arr[:] = np.nan
        for i, line in enumerate(lines):
            arr[i, : len(line)] = line
        return arr



from functools import lru_cache
from itertools import combinations, count
from typing import Any, Dict, List, Generator

from .reader import SpreadsheetReader, InvalidFileException

import numpy as np
import pyperclip

HELP = (
    "ARROWS / hjkl - Move",
    "<TAB> <SHIFT-TAB> / n <SHIFT n> - Switch sheet",
    "v - Visual/selection mode",
    "<ESC> - Exit visual/selection mode",
    "c / y - Copy selected cell(s)",
    "q - Exit",
    "? - Show this message",
)


def read_spreadsheet(
    filename: str, delimiter: str, fillna: Any, encoding: str
) -> Dict[str, np.ndarray]:
    return SpreadsheetReader(filename, delimiter, fillna, encoding).read()


@lru_cache()
def get_alphas(n: int) -> List[str]:
    alphaiter = get_alpha_iter()
    return [next(alphaiter) for _ in range(n)]


def copy(obj: Any):
    if isinstance(obj, np.ndarray):
        string = array_to_string(obj)
    else:
        string = str(obj)
    pyperclip.copy(string)


def get_alpha_iter() -> Generator[str, None, None]:
    alpha = range(65, 91)
    for i in count(1):
        chars = combinations(alpha, i)
        for ch in chars:
            yield "".join(map(chr, ch))


def array_to_string(mat: np.ndarray) -> str:
    return "\n".join(",".join(map(str, e)) for e in mat)



import curses
import math

import numpy as np
from typing import Callable, Dict, List, Optional

from .coordinate import Coordinate
from .keys import Keys, Key
from .styles import Styles
from . import utils


class Grid:
    def __init__(self, scr, sheets: Dict[str, np.ndarray], styles: Styles):
        self.scr = scr
        self.sheets: Dict[str, np.ndarray] = sheets
        self.sheetId: int = 0
        self.sheetNames: List[str] = list(self.sheets.keys())
        self.actions: Dict[Key, Callable] = self.get_actions()
        self.styles: Styles = styles
        self.init()

    def init(self):
        self.origin: Coordinate = Coordinate(0, 0)
        self.selected: Coordinate = Coordinate(0, 0)
        self.array: np.ndarray = self.sheets[self.sheetNames[self.sheetId]]
        self.visual: Optional[Coordinate] = None
        self.heigth, self.width = self.array.shape
        self.styles.init(self.heigth)
        self.draw()

    def redraw(self, func, *args, **kwargs) -> Callable:
        def runnable():
            func(*args, **kwargs)
            self.draw()

        return runnable

    def handle_press(self, key: int) -> bool:
        key = Keys.to_key(key)
        if key:
            if key in self.actions:
                self.actions[key]()
            elif key == Keys.QUIT:
                return False
        return True

    def get_actions(self) -> Dict[Key, Callable]:
        return {
            Keys.UP: self.redraw(self.move_vertical, -1),
            Keys.DOWN: self.redraw(self.move_vertical, 1),
            Keys.LEFT: self.redraw(self.move_horizontal, -1),
            Keys.RIGHT: self.redraw(self.move_horizontal, 1),
            Keys.TAB: self.redraw(self.switch_sheet, 1),
            Keys.SHIFT_TAB: self.redraw(self.switch_sheet, -1),
            Keys.VISUAL: self.redraw(self.set_visual),
            Keys.ESC: self.redraw(self.set_visual, reset=True),
            Keys.COPY: self.copy,
            Keys.HELP: self.help,
        }

    def set_visual(self, reset=False):
        self.visual = None if reset else Coordinate(self.selected.x, self.selected.y)

    def copy(self):
        if self.visual is None:
            utils.copy(self.array[self.selected.y, self.selected.x])
        else:
            ymin, ymax = sorted([self.selected.y, self.visual.y])
            xmin, xmax = sorted([self.selected.x, self.visual.x])
            utils.copy(self.array[ymin : ymax + 1, xmin : xmax + 1])
        self.visual = None
        self.draw()
        self.draw_footer("Copied")

    def help(self):
        self.scr.clear()
        self.scr.addstr(0, 0, "Help")
        for i, string in enumerate(utils.HELP, 2):
            if i >= self.scrheight:
                return
            self.scr.addstr(i, 1, string)

    def move_horizontal(self, direction: int):
        self.selected.x = (self.selected.x + direction) % self.width
        if (self.selected.x - self.origin.x) >= self.scrwidth_cells - 2:
            self.origin.x = self.selected.x - self.scrwidth_cells + 2
        elif self.selected.x < self.origin.x:
            self.origin.x = self.selected.x

    def move_vertical(self, direction: int):
        self.selected.y = (self.selected.y + direction) % self.heigth
        if (self.selected.y - self.origin.y) >= self.scrheight - 3:
            self.origin.y = self.selected.y - (self.scrheight - 3) + 1
        elif self.selected.y < self.origin.y:
            self.origin.y = self.selected.y

    def switch_sheet(self, direction: int):
        if len(self.sheetNames) > 1:
            self.sheetId = (self.sheetId + direction) % len(self.sheetNames)
            self.init()

    def set_screen_variables(self):
        self.scrheight, self.scrwidth = self.scr.getmaxyx()
        self.scrwidth_cells = math.ceil(
            (self.scrwidth - self.styles.lnwidth) / (self.styles.width + 1)
        )
        self.num_cols_in_screen = min(self.width - self.origin.x, self.scrwidth_cells)
        self.num_rows_in_screen = min(self.heigth - self.origin.y, self.scrheight - 3)

    def draw_header(self):
        alphas = utils.get_alphas(self.width)
        fmt = self.styles.get_header_format_string(self.num_cols_in_screen)
        header = fmt.format(
            *alphas[self.origin.x : self.origin.x + self.num_cols_in_screen]
        )
        self.scr.addstr(0, self.styles.lnwidth, header, self.styles.c_header)

    def get_cell_style(self, row: int, col: int):
        coord = self.origin + Coordinate(col, row)
        if self.visual is not None:
            if (
                coord.x >= min(self.selected.x, self.visual.x)
                and coord.x <= max(self.selected.x, self.visual.x)
                and coord.y >= min(self.selected.y, self.visual.y)
                and coord.y <= max(self.selected.y, self.visual.y)
            ):
                if coord == self.selected:
                    return self.styles.c_visual | curses.A_UNDERLINE
                return self.styles.c_visual
        if coord == self.selected:
            return self.styles.c_selection
        return (
            curses.A_UNDERLINE
            if (row == self.num_rows_in_screen - 1)
            else curses.A_NORMAL
        )

    def draw_grid(self):
        for row in range(self.num_rows_in_screen):
            self.scr.addstr(
                row + 1,
                0,
                self.styles.lineno_fmt.format(str(row + self.origin.y + 1)),
                self.styles.c_lineno,
            )
            for col in range(self.num_cols_in_screen):
                self.scr.addstr(
                    row + 1,
                    col * (self.styles.width + 1) + self.styles.lnwidth,
                    self.styles.format_cell(
                        self.array[row + self.origin.y, col + self.origin.x],
                    ),
                    self.get_cell_style(row, col),
                )
        self.scr.addstr(row + 2, 0, " " * self.scrwidth)

    def draw_sheets(self):
        sheetPosx = 0
        for i, sheet in enumerate(self.sheetNames):
            if i == self.sheetId:
                style = self.styles.c_sheet_selection
            else:
                style = self.styles.c_sheets
            self.scr.addstr(self.scrheight - 1, sheetPosx, sheet, style)
            sheetPosx += len(sheet) + 1

    def draw_footer(self, string=None):
        string = string or str(self.array[self.selected.y, self.selected.x])
        self.scr.addstr(
            self.scrheight - 2,
            0,
            self.styles.format_footer(string, self.scrwidth),
            self.styles.c_footer,
        )

    def draw(self):
        self.scr.clear()
        self.set_screen_variables()
        self.draw_header()
        self.draw_grid()
        self.draw_footer()
        self.draw_sheets()
